@@ -16,7 +16,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  * Copyright (C) 2007 - 2008 Novell, Inc.
- * Copyright (C) 2007 - 2008 Red Hat, Inc.
+ * Copyright (C) 2007 - 2010 Red Hat, Inc.
  */
 
 #include "config.h"
@@ -40,6 +40,7 @@
 #include "nm-dbus-manager.h"
 #include "nm-utils.h"
 #include "nm-device-wifi.h"
+#include "nm-device-olpc-mesh.h"
 #include "nm-device-ethernet.h"
 #include "nm-gsm-device.h"
 #include "nm-hso-gsm-device.h"
@@ -50,12 +51,26 @@
 
 #define HAL_DBUS_SERVICE "org.freedesktop.Hal"
 
+typedef struct {
+	GType device_type;
+	char *capability_str;
+	char *category;
+	gboolean (*is_device_fn) (NMHalManager *self, const char *udi);
+	NMDeviceCreatorFn creator_fn;
+} DeviceCreator;
+
+static void emit_udi_added (NMHalManager *self, const char *udi, DeviceCreator *creator);
+
 typedef struct {
 	LibHalContext *hal_ctx;
 	NMDBusManager *dbus_mgr;
 	GSList *device_creators;
 	gboolean rfkilled;  /* Authoritative rfkill state */
 
+	GSList *deferred_modems;
+	guint deferred_modem_id;
+	DeviceCreator *modem_creator;
+
 	/* Killswitch handling */
 	GSList *killswitch_list;
 	guint32 killswitch_poll_id;
@@ -63,6 +78,7 @@ typedef struct {
 	gboolean poll_rfkilled;
 	guint32 pending_polls;
 	GSList *poll_proxies;
+	gboolean killswitches_polled;
 
 	gboolean disposed;
 } NMHalManagerPrivate;
@@ -87,14 +103,6 @@ static gboolean poll_killswitches (gpointer user_data);
 
 /* Device creators */
 
-typedef struct {
-	GType device_type;
-	char *capability_str;
-	char *category;
-	gboolean (*is_device_fn) (NMHalManager *self, const char *udi);
-	NMDeviceCreatorFn creator_fn;
-} DeviceCreator;
-
 static DeviceCreator *
 get_creator (NMHalManager *self, const char *udi)
 {
@@ -118,18 +126,144 @@ get_creator (NMHalManager *self, const char *udi)
 /* Common helpers for built-in device creators */
 
 static char *
-nm_get_device_driver_name (LibHalContext *ctx, const char *origdev_udi)
+hal_get_subsystem (LibHalContext *ctx, const char *udi)
 {
-	char *driver_name = NULL;
+	char *subsys;
 
-	if (origdev_udi && libhal_device_property_exists (ctx, origdev_udi, "info.linux.driver", NULL)) {
-		char *drv = libhal_device_get_property_string (ctx, origdev_udi, "info.linux.driver", NULL);
+	subsys = libhal_device_get_property_string (ctx, udi, "info.subsystem", NULL);
+	if (!subsys) {
+		/* info.bus is deprecated */
+		subsys = libhal_device_get_property_string (ctx, udi, "info.bus", NULL);
+	}
+	return subsys;
+}
+
+static char *
+nm_get_device_driver_name (LibHalContext *ctx, const char *udi)
+{
+	char *driver_name = NULL, *drv, *tmp_udi, *tmp2;
+	int i = 1;
+
+	g_return_val_if_fail (udi != NULL, NULL);
+
+	/* get the UDI again so we can free it consistently with libhal_free_string */
+	tmp_udi = libhal_device_get_property_string (ctx, udi, "info.udi", NULL);
+
+	/* Look at up to 4 ancestors until we get a usable driver */
+	do {
+		drv = libhal_device_get_property_string (ctx, tmp_udi, "info.linux.driver", NULL);
+		if (drv) {
+			/* Check if we've gone too far */
+			if (   !strcmp (drv, "usb")
+			    || !strcmp (drv, "ehci_hcd")
+			    || !strcmp (drv, "yenta_cardbus")) {
+				/* Some drivers (libertas) are broken and report 'usb' as the
+				 * driver name even though the sysfs links are present.  Since
+				 * 0.7.2 used to handle that, we should stay compatible.
+				 */
+				if (strcmp (drv, "usb") != 0) {
+					libhal_free_string (drv);
+					drv = NULL;
+				}
+				break;
+			}
+		} else {
+			tmp2 = libhal_device_get_property_string (ctx, tmp_udi, "info.parent", NULL);
+			libhal_free_string (tmp_udi);
+			tmp_udi = tmp2;
+
+			/* Check if we've gone too far */
+			if (tmp_udi && !strcmp (tmp_udi, "/org/freedesktop/Hal/devices/computer"))
+				break;
+		}
+	} while (!drv && (i++ < 4) && tmp_udi);
+
+	libhal_free_string (tmp_udi);
+
+	if (drv) {
 		driver_name = g_strdup (drv);
 		libhal_free_string (drv);
 	}
+
 	return driver_name;
 }
 
+static gboolean
+get_vid_pid (LibHalContext *ctx,
+             const char *udi,
+             guint16 *out_vid,
+             guint16 *out_pid,
+             char **out_parent)
+{
+	int tmp;
+	char *parent;
+	DBusError error;
+	gboolean success = FALSE;
+
+	g_return_val_if_fail (ctx != NULL, FALSE);
+	g_return_val_if_fail (udi != NULL, FALSE);
+
+	parent = libhal_device_get_property_string (ctx, udi, "info.parent", NULL);
+	if (!parent)
+		return FALSE;
+
+	dbus_error_init (&error);
+	tmp = libhal_device_get_property_int (ctx, parent, "usb.vendor_id", &error);
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		goto out;
+	}
+	if (tmp < 0 || tmp > G_MAXUINT16)
+		goto out;
+	if (out_vid)
+		*out_vid = (guint16) tmp;
+
+	dbus_error_init (&error);
+	tmp = libhal_device_get_property_int (ctx, parent, "usb.product_id", &error);
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		goto out;
+	}
+	if (tmp < 0 || tmp > G_MAXUINT16)
+		goto out;
+	if (out_pid)
+		*out_pid = (guint16) tmp;
+
+	if (out_parent)
+		*out_parent = g_strdup (parent);
+	success = TRUE;
+
+out:
+	libhal_free_string (parent);
+	return success;
+}
+
+/* Returns the parent if the device is a Sony Ericsson 'mbm'-style device */
+static gboolean
+is_mbm (LibHalContext *ctx, const char *udi, char **out_parent)
+{
+	guint16 vendor_id = 0, product_id = 0;
+	char *parent = NULL;
+
+	if (!get_vid_pid (ctx, udi, &vendor_id, &product_id, &parent))
+		return FALSE;
+
+	if (   (vendor_id == 0x0bdb && product_id == 0x1900)    /* SE F3507g */
+	    || (vendor_id == 0x0bdb && product_id == 0x1902)    /* SE F3507g */
+	    || (vendor_id == 0x0bdb && product_id == 0x1904)    /* SE F3607gw */
+	    || (vendor_id == 0x0bdb && product_id == 0x1906)    /* SE F3307 */
+	    || (vendor_id == 0x0fce && product_id == 0xd0cf)    /* SE MD300 */
+	    || (vendor_id == 0x413c && product_id == 0x8147)    /* Dell 5530 HSDPA */
+	    || (vendor_id == 0x0930 && product_id == 0x130b)) { /* Toshiba */
+		if (out_parent)
+			*out_parent = parent;
+		return TRUE;
+	}
+
+	g_free (parent);
+	return FALSE;
+}
+
 /* Wired device creator */
 
 static gboolean
@@ -159,9 +293,8 @@ wired_device_creator (NMHalManager *self,
                       gboolean managed)
 {
 	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
-	GObject *device;
-	char *iface;
-	char *driver;
+	GObject *device = NULL;
+	char *iface, *driver;
 
 	iface = libhal_device_get_property_string (priv->hal_ctx, udi, "net.interface", NULL);
 	if (!iface) {
@@ -169,12 +302,16 @@ wired_device_creator (NMHalManager *self,
 		return NULL;
 	}
 
-	driver = nm_get_device_driver_name (priv->hal_ctx, origdev_udi);
-	device = (GObject *) nm_device_ethernet_new (udi, iface, driver, managed);
+	driver = nm_get_device_driver_name (priv->hal_ctx, udi);
+
+	/* Special handling of Ericsson F3507g 'mbm' devices; ignore the
+	 * cdc-ether device that it provides since we don't use it yet.
+	 */
+	if (!is_mbm (priv->hal_ctx, udi, NULL))
+		device = (GObject *) nm_device_ethernet_new (udi, iface, driver, managed);
 
 	libhal_free_string (iface);
 	g_free (driver);
-
 	return device;
 }
 
@@ -217,7 +354,7 @@ wireless_device_creator (NMHalManager *self,
 		return NULL;
 	}
 
-	driver = nm_get_device_driver_name (priv->hal_ctx, origdev_udi);
+	driver = nm_get_device_driver_name (priv->hal_ctx, udi);
 	device = (GObject *) nm_device_wifi_new (udi, iface, driver, managed);
 
 	libhal_free_string (iface);
@@ -247,60 +384,106 @@ is_modem_device (NMHalManager *self, const char *udi)
 	return is_modem;
 }
 
+static char *
+hal_get_originating_device (LibHalContext *ctx, const char *udi, const char *prefix)
+{
+	char *od_key = g_strdup_printf ("%s.originating_device", prefix);
+	char *pd_key = g_strdup_printf ("%s.physical_device", prefix);
+	char *od;
+
+	od = libhal_device_get_property_string (ctx, udi, od_key, NULL);
+	if (!od) {
+		/* physical_device is deprecated */
+		od = libhal_device_get_property_string (ctx, udi, pd_key, NULL);
+	}
+
+	g_free (od_key);
+	g_free (pd_key);
+	return od;
+}
+
 static char *
 get_hso_netdev (LibHalContext *ctx, const char *udi)
 {
-	char *serial_parent, *netdev = NULL;
-	char **netdevs;
+	char *serial_od = NULL, *serial_od_parent = NULL, *netdev = NULL, *subsys;
+	char **netdevs = NULL;
 	int num, i;
 
 	/* Get the serial interface's originating device UDI, used to find the
 	 * originating device's netdev.
 	 */
-	serial_parent = libhal_device_get_property_string (ctx, udi, "serial.originating_device", NULL);
-	if (!serial_parent)
-		serial_parent = libhal_device_get_property_string (ctx, udi, "info.parent", NULL);
-	if (!serial_parent)
-		return NULL;
+	serial_od = hal_get_originating_device (ctx, udi, "serial");
+	if (!serial_od)
+		goto out;
+
+	serial_od_parent = libhal_device_get_property_string (ctx, serial_od, "info.parent", NULL);
+	if (!serial_od_parent)
+		goto out;
+
+	/* Check to ensure we've got the actual "USB Device" */
+	subsys = hal_get_subsystem (ctx, serial_od_parent);
+	if (!subsys || strcmp (subsys, "usb_device")) {
+		libhal_free_string (subsys);
+		goto out;
+	}
+	libhal_free_string (subsys);
 
 	/* Look for the originating device's netdev */
 	netdevs = libhal_find_device_by_capability (ctx, "net", &num, NULL);
 	for (i = 0; netdevs && !netdev && (i < num); i++) {
-		char *netdev_parent, *tmp;
+		char *net_od = NULL, *net_od_parent = NULL, *tmp;
 
-		netdev_parent = libhal_device_get_property_string (ctx, netdevs[i], "net.originating_device", NULL);
-		if (!netdev_parent)
-			netdev_parent = libhal_device_get_property_string (ctx, netdevs[i], "net.physical_device", NULL);
-		if (!netdev_parent)
-			continue;
+		net_od = hal_get_originating_device (ctx, netdevs[i], "net");
+		if (!net_od)
+			goto next;
 
-		if (!strcmp (netdev_parent, serial_parent)) {
+		net_od_parent = libhal_device_get_property_string (ctx, net_od, "info.parent", NULL);
+		if (!net_od_parent)
+			goto next;
+
+		/* Check to ensure we've got the actual "USB Device" */
+		subsys = hal_get_subsystem (ctx, net_od_parent);
+		if (!subsys || strcmp (subsys, "usb_device")) {
+			libhal_free_string (subsys);
+			goto next;
+		}
+		libhal_free_string (subsys);
+
+		if (!strcmp (net_od_parent, serial_od_parent)) {
 			/* We found it */
 			tmp = libhal_device_get_property_string (ctx, netdevs[i], "net.interface", NULL);
-			if (tmp) {
+			if (tmp)
 				netdev = g_strdup (tmp);
-				libhal_free_string (tmp);
-			}
+			libhal_free_string (tmp);
 		}
 
-		libhal_free_string (netdev_parent);
+	next:
+		libhal_free_string (net_od_parent);
+		libhal_free_string (net_od);
 	}
+
+out:
 	libhal_free_string_array (netdevs);
-	libhal_free_string (serial_parent);
+	libhal_free_string (serial_od);
+	libhal_free_string (serial_od_parent);
 
 	return netdev;
 }
 
-#define PROP_GSM   "ID_NM_MODEM_GSM"
-#define PROP_CDMA  "ID_NM_MODEM_IS707_A"
-#define PROP_EVDO1 "ID_NM_MODEM_IS856"
-#define PROP_EVDOA "ID_NM_MODEM_IS856_A"
+#define PROP_GSM      "ID_NM_MODEM_GSM"
+#define PROP_CDMA     "ID_NM_MODEM_IS707_A"
+#define PROP_EVDO1    "ID_NM_MODEM_IS856"
+#define PROP_EVDOA    "ID_NM_MODEM_IS856_A"
+#define PROP_ZTE_MAIN "ID_NM_ZTE_PORT_TYPE_MODEM"
+#define PROP_ZTE_SET  "ID_NM_ZTE_PORT_TYPE_SET"
 
 #if HAVE_LIBUDEV
 
 typedef struct {
 	gboolean gsm;
 	gboolean cdma;
+	gboolean zte_main;
+	gboolean zte_set;
 } UdevIterData;
 
 #if UDEV_VERSION >= 129
@@ -333,6 +516,13 @@ static int udev_device_prop_iter(struct udev_device *udev_device,
 	if (!strcmp (key, PROP_EVDOA) && !strcmp (value, "1"))
 		types->cdma = TRUE;
 
+	if (types->cdma || types->gsm) {
+	    if (!strcmp (key, PROP_ZTE_MAIN) && !strcmp (value, "1"))
+			types->zte_main = TRUE;
+	    if (!strcmp (key, PROP_ZTE_SET) && !strcmp (value, "1"))
+			types->zte_set = TRUE;
+	}
+
 	/* Return 0 to continue looking */
 	return types->gsm && types->cdma;
 }
@@ -341,7 +531,9 @@ static int udev_device_prop_iter(struct udev_device *udev_device,
 static gboolean
 libudev_get_modem_capabilities (const char *sysfs_path,
                                 gboolean *gsm,
-                                gboolean *cdma)
+                                gboolean *cdma,
+                                gboolean *zte_main,
+                                gboolean *zte_set)
 {
 	struct udev *udev;
 	struct udev_device *device;
@@ -351,6 +543,10 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 	g_return_val_if_fail (*gsm == FALSE, FALSE);
 	g_return_val_if_fail (cdma != NULL, FALSE);
 	g_return_val_if_fail (*cdma == FALSE, FALSE);
+	g_return_val_if_fail (zte_main != NULL, FALSE);
+	g_return_val_if_fail (*zte_main == FALSE, FALSE);
+	g_return_val_if_fail (zte_set != NULL, FALSE);
+	g_return_val_if_fail (*zte_set == FALSE, FALSE);
 
 	udev = udev_new ();
 	if (!udev)
@@ -378,6 +574,8 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 		const char *cdma_val = get_udev_property (device, PROP_CDMA);
 		const char *evdo1_val = get_udev_property (device, PROP_EVDO1);
 		const char *evdoa_val = get_udev_property (device, PROP_EVDOA);
+		const char *zte_main_val = get_udev_property (device, PROP_ZTE_MAIN);
+		const char *zte_set_val = get_udev_property (device, PROP_ZTE_SET);
 
 		if (gsm_val && !strcmp (gsm_val, "1"))
 			*gsm = TRUE;
@@ -387,6 +585,13 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 			*cdma = TRUE;
 		if (evdoa_val && !strcmp (evdoa_val, "1"))
 			*cdma = TRUE;
+
+		if (*gsm || *cdma) {
+		    if (zte_main_val && !strcmp (zte_main_val, "1"))
+				*zte_main = TRUE;
+		    if (zte_set_val && !strcmp (zte_set_val, "1"))
+				*zte_set = TRUE;
+		}
 	}
 #else
 	{
@@ -395,6 +600,10 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 		udev_device_get_properties (device, udev_device_prop_iter, &iterdata);
 		*gsm = iterdata.gsm;
 		*cdma = iterdata.cdma;
+		if (*gsm || *cdma) {
+			*zte_main = iterdata.zte_main;
+			*zte_set = iterdata.zte_set;
+		}
 	}
 #endif
 
@@ -406,7 +615,9 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 static gboolean
 udevadm_get_modem_capabilities (const char *sysfs_path,
                                 gboolean *gsm,
-                                gboolean *cdma)
+                                gboolean *cdma,
+                                gboolean *zte_main,
+                                gboolean *zte_set)
 {
 	char *udevadm_argv[] = { "/sbin/udevadm", "info", "--query=env", NULL, NULL };
 	char *syspath_arg = NULL;
@@ -414,13 +625,17 @@ udevadm_get_modem_capabilities (const char *sysfs_path,
 	int exitcode;
 	GError *error = NULL;
 	char **lines = NULL, **iter;
-	gboolean success = FALSE;
+	gboolean success = FALSE, is_zte_main = FALSE, is_zte_set = FALSE;
 
 	g_return_val_if_fail (sysfs_path != NULL, FALSE);
 	g_return_val_if_fail (gsm != NULL, FALSE);
 	g_return_val_if_fail (*gsm == FALSE, FALSE);
 	g_return_val_if_fail (cdma != NULL, FALSE);
 	g_return_val_if_fail (*cdma == FALSE, FALSE);
+	g_return_val_if_fail (zte_main != NULL, FALSE);
+	g_return_val_if_fail (*zte_main == FALSE, FALSE);
+	g_return_val_if_fail (zte_set != NULL, FALSE);
+	g_return_val_if_fail (*zte_set == FALSE, FALSE);
 
 	udevadm_argv[3] = syspath_arg = g_strdup_printf ("--path=%s", sysfs_path);
 	if (g_spawn_sync ("/", udevadm_argv, NULL, 0, NULL, NULL,
@@ -445,14 +660,21 @@ udevadm_get_modem_capabilities (const char *sysfs_path,
 	for (iter = lines; *iter; iter++) {
 		if (!strcmp (*iter, PROP_GSM "=1")) {
 			*gsm = TRUE;
-			break;
 		} else if (   !strcmp (*iter, PROP_CDMA "=1")
 		           || !strcmp (*iter, PROP_EVDO1 "=1")
 		           || !strcmp (*iter, PROP_EVDOA "=1")) {
 			*cdma = TRUE;
-			break;
-		}
+		} else if (!strcmp (*iter, PROP_ZTE_MAIN "=1"))
+			is_zte_main = TRUE;
+		else if (!strcmp (*iter, PROP_ZTE_SET "=1"))
+			is_zte_set = TRUE;
 	}
+
+	if (*gsm || *cdma) {
+		*zte_main = is_zte_main;
+		*zte_set = is_zte_set;
+	}
+
 	success = TRUE;
 
 error:
@@ -464,7 +686,7 @@ error:
 }
 #endif
 
-static gboolean
+static void
 hal_get_modem_capabilities (LibHalContext *ctx,
                             const char *udi,
                             gboolean *gsm,
@@ -472,16 +694,16 @@ hal_get_modem_capabilities (LibHalContext *ctx,
 {
 	char **capabilities, **iter;
 
-	g_return_val_if_fail (ctx != NULL, FALSE);
-	g_return_val_if_fail (udi != NULL, FALSE);
-	g_return_val_if_fail (gsm != NULL, FALSE);
-	g_return_val_if_fail (*gsm == FALSE, FALSE);
-	g_return_val_if_fail (cdma != NULL, FALSE);
-	g_return_val_if_fail (*cdma == FALSE, FALSE);
+	g_return_if_fail (ctx != NULL);
+	g_return_if_fail (udi != NULL);
+	g_return_if_fail (gsm != NULL);
+	g_return_if_fail (*gsm == FALSE);
+	g_return_if_fail (cdma != NULL);
+	g_return_if_fail (*cdma == FALSE);
 
 	/* Make sure it has the 'modem' capability first */
 	if (!libhal_device_query_capability (ctx, udi, "modem", NULL))
-		return TRUE;
+		return;
 
 	capabilities = libhal_device_get_property_strlist (ctx, udi, "modem.command_sets", NULL);
 	/* 'capabilites' may be NULL */
@@ -497,81 +719,207 @@ hal_get_modem_capabilities (LibHalContext *ctx,
 	}
 	if (capabilities)
 		g_strfreev (capabilities);
+}
 
-	/* Compatiblity with the pre-specification bits */
-	if (!*gsm && !*cdma) {
-		capabilities = libhal_device_get_property_strlist (ctx, udi, "info.capabilities", NULL);
-		for (iter = capabilities; iter && *iter; iter++) {
-			if (!strcmp (*iter, "gsm")) {
-				*gsm = TRUE;
-				break;
-			}
-			if (!strcmp (*iter, "cdma")) {
-				*cdma = TRUE;
-				break;
-			}
-		}
-		if (capabilities)
-			g_strfreev (capabilities);
+typedef struct {
+	guint32 refcount;
+	char *udi;
+	char *origdev_udi;
+	gboolean gsm;
+	gboolean cdma;
+	guint32 usb_interface_number;
+	guint32 serial_port;
+} DeferredModem;
+
+#if 0
+static void
+deferred_modem_ref (DeferredModem *modem)
+{
+	g_return_if_fail (modem != NULL);
+	g_return_if_fail (modem->refcount > 0);
+	modem->refcount++;
+}
+#endif
+
+static void
+deferred_modem_unref (DeferredModem *modem)
+{
+	g_return_if_fail (modem != NULL);
+	g_return_if_fail (modem->refcount > 0);
+
+	modem->refcount--;
+	if (modem->refcount == 0) {
+		g_free (modem->udi);
+		g_free (modem->origdev_udi);
+		memset (modem, 0, sizeof (DeferredModem));
+		g_free (modem);
+	}
+}
+
+static DeferredModem *
+deferred_modem_new (const char *udi,
+                    const char *origdev_udi,
+                    gboolean gsm,
+                    gboolean cdma,
+                    LibHalContext *ctx)
+{
+	DeferredModem *modem;
+	char *parent;
+	dbus_int32_t serial_port = -1, usb_interface = -1;
+	DBusError error;
+
+	parent = libhal_device_get_property_string (ctx, udi, "info.parent", NULL);
+	if (!parent) {
+		g_warning ("couldn't get parent for '%s'", udi);
+		return NULL;
 	}
 
-	return TRUE;
+	modem = g_new0 (DeferredModem, 1);
+	modem->refcount = 1;
+	modem->udi = g_strdup (udi);
+	modem->origdev_udi = g_strdup (origdev_udi);
+	modem->gsm = gsm;
+	modem->cdma = cdma;
+
+	dbus_error_init (&error);
+	usb_interface = libhal_device_get_property_int (ctx, parent, "usb.interface.number", &error);
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		usb_interface = -1;
+	}
+
+	dbus_error_init (&error);
+	serial_port = libhal_device_get_property_int (ctx, udi, "serial.port", &error);
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		serial_port = -1;
+	}
+
+	if (usb_interface == -1 && serial_port == -1) {
+		g_warning ("couldn't get usb.interface.number and serial.port for '%s'", udi);
+		deferred_modem_unref (modem);
+		modem = NULL;
+	} else {
+		modem->usb_interface_number = usb_interface;
+		modem->serial_port = serial_port;
+	}
+
+	libhal_free_string (parent);
+	return modem;
+}
+
+static gint
+deferred_modem_sort_func (const DeferredModem *a, const DeferredModem *b)
+{
+	gint ret;
+
+	/* Sort is: origdev_udi, usb_interface, serial_port */
+
+	ret = strcmp (a->origdev_udi, b->origdev_udi);
+	if (ret)
+		return ret;
+
+	if (a->usb_interface_number < b->usb_interface_number)
+		return -1;
+	else if (a->usb_interface_number > b->usb_interface_number)
+		return 1;
+
+	if (a->serial_port < b->serial_port)
+		return -1;
+	else if (a->serial_port > b->serial_port)
+		return 1;
+
+	return 0;
+}
+
+static DeferredModem *
+deferred_modem_find (NMHalManager *self, const char *udi)
+{
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	GSList *iter;
+
+	for (iter = priv->deferred_modems; iter; iter = g_slist_next (iter)) {
+		DeferredModem *candidate = iter->data;
+
+		if (!strcmp (candidate->udi, udi))
+			return candidate;
+	}
+	return NULL;
+}
+
+static DeferredModem *
+deferred_modem_find_by_origdev (NMHalManager *self, const char *origdev_udi)
+{
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	GSList *iter;
+
+	for (iter = priv->deferred_modems; iter; iter = g_slist_next (iter)) {
+		DeferredModem *candidate = iter->data;
+
+		if (!strcmp (candidate->origdev_udi, origdev_udi))
+			return candidate;
+	}
+	return NULL;
+}
+
+static void
+deferred_modem_remove (NMHalManager *self, DeferredModem *modem)
+{
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	GSList *iter, *new = NULL;
+
+	g_return_if_fail (modem != NULL);
+
+	for (iter = priv->deferred_modems; iter; iter = g_slist_next (iter)) {
+		if (iter->data != modem)
+			new = g_slist_insert_sorted (new, iter->data, (GCompareFunc) deferred_modem_sort_func);
+	}
+	g_slist_free (priv->deferred_modems);
+	priv->deferred_modems = new;
+}
+
+static gboolean
+deferred_modem_timeout (gpointer data)
+{
+	NMHalManager *self = data;
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	GSList *iter;
+
+	nm_info ("Re-checking deferred serial ports");
+	for (iter = priv->deferred_modems; iter; iter = g_slist_next (iter)) {
+		DeferredModem *modem = iter->data;
+
+		/* re-emit udi-added for each UDI that was deferred */
+		emit_udi_added (self, modem->udi, priv->modem_creator);
+	}
+
+	g_slist_foreach (priv->deferred_modems, (GFunc) deferred_modem_unref, NULL);
+	g_slist_free (priv->deferred_modems);
+	priv->deferred_modems = NULL;
+	priv->deferred_modem_id = 0;
+	return FALSE;
 }
 
 static GObject *
-modem_device_creator (NMHalManager *self,
-                      const char *udi,
-                      const char *origdev_udi,
-                      gboolean managed)
+new_modem_device (const char *udi,
+                  gboolean gsm,
+                  gboolean cdma,
+                  const char *ttyname,
+                  const char *driver,
+                  gboolean managed,
+                  LibHalContext *ctx)
 {
-	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
-	char *serial_device;
-	const char *ttyname;
-	char *sysfs_path;
-	char *driver = NULL;
 	GObject *device = NULL;
-	gboolean type_gsm = FALSE;
-	gboolean type_cdma = FALSE;
 	char *netdev = NULL;
-	gboolean udev = TRUE;
 
-	serial_device = libhal_device_get_property_string (priv->hal_ctx, udi, "serial.device", NULL);
-	driver = nm_get_device_driver_name (priv->hal_ctx, origdev_udi);
-	if (!serial_device || !driver)
-		goto out;
+	g_return_val_if_fail (udi != NULL, NULL);
+	g_return_val_if_fail (gsm || cdma, NULL);
+	g_return_val_if_fail (ttyname != NULL, NULL);
+	g_return_val_if_fail (driver != NULL, NULL);
+	g_return_val_if_fail (ctx != NULL, NULL);
 
-	/* Try udev first */
-	sysfs_path = libhal_device_get_property_string (priv->hal_ctx, udi, "linux.sysfs_path", NULL);
-	if (!sysfs_path) {
-		nm_warning ("could not determine sysfs path for '%s'", serial_device);
-		goto out;
-	}
-
-	ttyname = serial_device + strlen ("/dev/");
-
-#if HAVE_LIBUDEV
-	libudev_get_modem_capabilities (sysfs_path, &type_gsm, &type_cdma);
-#else
-	udevadm_get_modem_capabilities (sysfs_path, &type_gsm, &type_cdma);
-#endif
-	libhal_free_string (sysfs_path);
-
-	/* If udev didn't know anything, try deprecated HAL modem capabilities */
-	if (!type_gsm && !type_cdma) {
-		hal_get_modem_capabilities (priv->hal_ctx, udi, &type_gsm, &type_cdma);
-		udev = FALSE;
-	}
-
-	if (!type_gsm && !type_cdma)
-		goto out;
-
-	nm_info ("(%s): detected %s modem via %s capabilities",
-	         ttyname,
-	         type_gsm ? "GSM" : "CDMA",
-	         udev ? "udev" : "HAL");
-
-	/* Special handling of 'hso' cards (until punted out to a modem manager) */
-	if (type_gsm && !strcmp (driver, "hso")) {
+	/* Special handling of 'hso' cards (until punted to ModemManager) */
+	if (gsm && !strcmp (driver, "hso")) {
 		char *hsotype_path;
 		char *contents = NULL;
 		gboolean success;
@@ -587,26 +935,254 @@ modem_device_creator (NMHalManager *self,
 			if (   !strstr (contents, "Control")
 			    && !strstr (contents, "control")) {
 				g_free (contents);
-				goto out;
+				return NULL;
 			}
 			g_free (contents);
 		}
 
-		netdev = get_hso_netdev (priv->hal_ctx, udi);
+		netdev = get_hso_netdev (ctx, udi);
+		if (!netdev) {
+			nm_warning ("couldn't find HSO modem network device.");
+			return NULL;
+		}
 	}
 
-	if (type_gsm) {
-		if (netdev)
-			device = (GObject *) nm_hso_gsm_device_new (udi, ttyname, NULL, netdev, driver, managed);
-		else
-			device = (GObject *) nm_gsm_device_new (udi, ttyname, NULL, driver, managed);
-	} else if (type_cdma)
+	/* Special handling of Option cards (until punted to ModemManager).  Only
+	 * the first USB interface can be used for control and PPP.
+	 */
+	if (gsm && !strcmp (driver, "option")) {
+		char *parent;
+		guint32 vendor_id = 0, usb_interface = 0;
+
+		parent = libhal_device_get_property_string (ctx, udi, "info.parent", NULL);
+		if (!parent)
+			return NULL;
+
+		vendor_id = libhal_device_get_property_int (ctx, parent, "usb.vendor_id", NULL);
+		if (vendor_id == 0x0AF0) {
+			usb_interface = libhal_device_get_property_int (ctx, parent, "usb.interface.number", NULL);
+			if (usb_interface > 0) {
+				libhal_free_string (parent);
+				return NULL;
+			}
+		}
+		libhal_free_string (parent);
+	}
+
+	/* Special handling of Ericsson F3507g 'mbm' devices; already handled by
+	 * ModemManager more correctly in HEAD.
+	 */
+	if (gsm && !strcmp (driver, "cdc_acm")) {
+		char *parent = NULL;
+		guint32 usb_interface;
+
+		if (is_mbm (ctx, udi, &parent)) {
+			usb_interface = libhal_device_get_property_int (ctx, parent, "usb.interface.number", NULL);
+			g_free (parent);
+			if (usb_interface != 1)
+				return NULL;
+		}
+	}
+
+	if (gsm && netdev) {
+		device = (GObject *) nm_hso_gsm_device_new (udi, ttyname, NULL, netdev, driver, managed);
+		g_free (netdev);
+	} else if (gsm)
+		device = (GObject *) nm_gsm_device_new (udi, ttyname, NULL, driver, managed);
+	else if (cdma)
 		device = (GObject *) nm_cdma_device_new (udi, ttyname, NULL, driver, managed);
+	else
+		g_assert_not_reached ();
+
+	return device;
+}
+
+static GObject *
+modem_device_creator (NMHalManager *self,
+                      const char *udi,
+                      const char *origdev_udi,
+                      gboolean managed)
+{
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	char *serial_device;
+	const char *ttyname;
+	char *sysfs_path;
+	char *driver = NULL;
+	GObject *device = NULL;
+	gboolean udev_gsm = FALSE;
+	gboolean udev_cdma = FALSE;
+	gboolean hal_gsm = FALSE;
+	gboolean hal_cdma = FALSE;
+	gboolean later = FALSE;
+	gboolean udev_success = FALSE;
+	gboolean zte_main = FALSE;
+	gboolean zte_set = FALSE;
+	DeferredModem *deferred = NULL;
+	guint16 vid = 0;
+
+	serial_device = libhal_device_get_property_string (priv->hal_ctx, udi, "serial.device", NULL);
+	/* For serial devices, 'origdev_udi' will be the actual USB or platform device,
+	 * not HAL's 'serial.originating_device'.
+	 */
+	driver = nm_get_device_driver_name (priv->hal_ctx, udi);
+	if (!serial_device || !driver)
+		goto out;
+
+	ttyname = serial_device + strlen ("/dev/");
+
+	/* If the UDI was deferred from earlier, just make the device with the
+	 * cached capabilities.
+	 */
+	deferred = deferred_modem_find (self, udi);
+	if (deferred) {
+		udev_gsm = deferred->gsm;
+		udev_cdma = deferred->cdma;
+		device = new_modem_device (udi, deferred->gsm, deferred->cdma, ttyname, driver, managed, priv->hal_ctx);
+		goto out;
+	}
+
+	/* Get udev probed capabilities */
+	sysfs_path = libhal_device_get_property_string (priv->hal_ctx, udi, "linux.sysfs_path", NULL);
+	if (!sysfs_path) {
+		nm_warning ("(%s): could not determine sysfs path", ttyname);
+		goto out;
+	}
+
+#if HAVE_LIBUDEV
+	udev_success = libudev_get_modem_capabilities (sysfs_path, &udev_gsm, &udev_cdma, &zte_main, &zte_set);
+#else
+	udev_success = udevadm_get_modem_capabilities (sysfs_path, &udev_gsm, &udev_cdma, &zte_main, &zte_set);
+#endif
+	libhal_free_string (sysfs_path);
+
+	/* Get HAL capabilities too */
+	hal_get_modem_capabilities (priv->hal_ctx, udi, &hal_gsm, &hal_cdma);
+
+	/* If for some reason running the udev prober failed fall back to HAL */
+	if (!udev_success) {
+		nm_warning ("(%s): udev probing failed; using only HAL modem capabilities.", ttyname);
+		udev_gsm = hal_gsm;
+		udev_cdma = hal_cdma;
+	}
+
+	/* ZTE modem ports get tagged by udev rules, so if the modem was known to
+	 * the rules (in which case ID_NM_ZTE_PORT_TYPE_SET will be set), and the
+	 * port isn't a "main" port, we should ignore it.
+	 */
+	if (get_vid_pid (priv->hal_ctx, udi, &vid, NULL, NULL) && (vid == 0x19d2)) {
+		if (zte_set && !zte_main) {
+			udev_gsm = FALSE;
+			udev_cdma = FALSE;
+		}
+	}
+
+	/* If it's not known to either udev or HAL as a modem, nothing to do */
+	if (!udev_gsm && !udev_cdma && !hal_gsm && !hal_cdma) {
+		nm_info ("(%s): ignoring due to lack of mobile broadband capabilties", ttyname);
+		goto out;
+	}
+
+	nm_info ("(%s): found serial port (udev:%s%s%s  hal:%s%s%s)",
+	         ttyname,
+	         udev_gsm ? "GSM" : "", udev_gsm && udev_cdma ? "/" : "", udev_cdma ? "CDMA" : "",
+	         hal_gsm ? "GSM" : "", hal_gsm && hal_cdma ? "/" : "", hal_cdma ? "CDMA" : "");
+
+	/* If other ports owned by this port's originating device are already
+	 * deferred, defer this port as well.
+	 */
+	later = !!deferred_modem_find_by_origdev (self, origdev_udi);
+
+	/* The logic goes like this:
+	 *
+	 * a) if both udev and HAL agree on the port command sets, use it
+	 * b) if HAL thinks its a modem, but the command set disagrees with udev,
+	 *      assume udev probing is correct
+	 * c) if HAL thinks it's a modem but udev does not, don't use it
+	 * d) if HAL doesn't think it's a modem but udev does, schedule a timer
+	 *      to check back later after HAL is done sending ttys.
+	 */
+	if (udev_gsm == hal_gsm && udev_cdma == hal_cdma) {
+		/* Case (a): HAL and udev agree. */
+		if (!later)
+			device = new_modem_device (udi, udev_gsm, udev_cdma, ttyname, driver, managed, priv->hal_ctx);
+	} else if ((hal_gsm || hal_cdma) && (udev_gsm || udev_cdma)) {
+		/* Case (b): HAL and udev think it's a modem, but they don't
+		 * agree on the command set the modem supports.  Trust udev.
+		 */
+		if (!later)
+			device = new_modem_device (udi, udev_gsm, udev_cdma, ttyname, driver, managed, priv->hal_ctx);
+	} else if ((hal_gsm || hal_cdma) && !udev_gsm && !udev_cdma) {
+		/* Case (c): HAL thinks it's a modem, but udev doesn't */
+		nm_info ("(%s): ignoring due to lack of probed mobile broadband capabilties", ttyname);
+		later = FALSE;
+	} else {
+		/* Case (d): HAL doesn't think its a modem, but udev does */
+		later = TRUE;
+	}
+
+	if (!device && later) {
+		deferred = deferred_modem_new (udi, origdev_udi, udev_gsm, udev_cdma, priv->hal_ctx);
+		priv->deferred_modems = g_slist_insert_sorted (priv->deferred_modems,
+		                                               deferred,
+		                                               (GCompareFunc) deferred_modem_sort_func);
+
+		if (priv->deferred_modem_id)
+			g_source_remove (priv->deferred_modem_id);
+		priv->deferred_modem_id = g_timeout_add_seconds (8, deferred_modem_timeout, self);
+
+		nm_info ("(%s): deferring until all ports found", ttyname);
+	}
 
 out:
 	libhal_free_string (serial_device);
 	g_free (driver);
+	return device;
+}
 
+/* OLPC mesh creator */
+static gboolean
+is_olpc_mesh_device (NMHalManager *self, const char *udi)
+{
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	gboolean ret = FALSE;
+
+	nm_debug ("Checking if %s is an olpc mesh device", udi);
+
+	if (libhal_device_property_exists (priv->hal_ctx, udi, "info.category", NULL)) {
+		char *category;
+
+		category = libhal_device_get_property_string (priv->hal_ctx, udi, "info.category", NULL);
+		if (category) {
+			ret = strcmp (category, "net.olpcmesh") == 0;
+			libhal_free_string (category);
+		}
+	}
+
+	return ret;
+}
+
+static GObject *
+olpc_mesh_device_creator (NMHalManager *self,
+						  const char *udi,
+						  const char *origdev_udi,
+						  gboolean managed)
+{
+	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
+	GObject *device;
+	char *iface;
+	char *driver;
+
+	iface = libhal_device_get_property_string (priv->hal_ctx, udi, "net.interface", NULL);
+	if (!iface) {
+		nm_warning ("Couldn't get interface for %s, ignoring.", udi);
+		return NULL;
+	}
+
+	driver = nm_get_device_driver_name (priv->hal_ctx, udi);
+	device = (GObject *) nm_device_olpc_mesh_new (udi, iface, driver, managed);
+
+	libhal_free_string (iface);
+	g_free (driver);
 	return device;
 }
 
@@ -634,6 +1210,15 @@ register_built_in_creators (NMHalManager *self)
 	creator->creator_fn = wireless_device_creator;
 	priv->device_creators = g_slist_append (priv->device_creators, creator);
 
+	/* OLPC mesh device */
+	creator = g_slice_new0 (DeviceCreator);
+	creator->device_type = NM_TYPE_DEVICE_OLPC_MESH;
+	creator->capability_str = g_strdup ("net.olpcmesh");
+	creator->category = g_strdup ("net");
+	creator->is_device_fn = is_olpc_mesh_device;
+	creator->creator_fn = olpc_mesh_device_creator;
+	priv->device_creators = g_slist_append (priv->device_creators, creator);
+
 	/* Modem */
 	creator = g_slice_new0 (DeviceCreator);
 	creator->device_type = NM_TYPE_SERIAL_DEVICE;
@@ -642,28 +1227,60 @@ register_built_in_creators (NMHalManager *self)
 	creator->is_device_fn = is_modem_device;
 	creator->creator_fn = modem_device_creator;
 	priv->device_creators = g_slist_append (priv->device_creators, creator);
+	priv->modem_creator = creator;
 }
 
 static void
 emit_udi_added (NMHalManager *self, const char *udi, DeviceCreator *creator)
 {
 	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
-	char *od, *tmp;
+	char *od = NULL, *tmp, *parent, *subsys = NULL;
 
 	g_return_if_fail (self != NULL);
 	g_return_if_fail (udi != NULL);
 	g_return_if_fail (creator != NULL);
 
-	tmp = g_strdup_printf ("%s.originating_device", creator->category);
-	od = libhal_device_get_property_string (priv->hal_ctx, udi, tmp, NULL);
-	g_free (tmp);
+	/* For USB serial devices, originating device should really be the "USB Device"
+	 * object, which is the grandparent of the tty device.  Only this grandparent
+	 * is really common among all ttys of the device; HAL's "serial.originating_device"
+	 * points to the "USB Interface" parent of the tty, but this interface only
+	 * sometimes has multiple child ttys.  More commonly, the "USB Device"
+	 * object provides several "USB Interface" objects, which each provide one tty.
+	 * Thus, to ensure that NM only uses the correct tty for the deivce,
+	 * we need to filter on the "USB Device" instead of the "USB Interface".
+	 */
+	parent = libhal_device_get_property_string (priv->hal_ctx, udi, "info.parent", NULL);
+	if (parent)
+		subsys = hal_get_subsystem (priv->hal_ctx, parent);
+	if (subsys && !strcmp (subsys, "usb")) {
+		char *usb_intf_udi;
 
-	if (!od) {
-		/* Older HAL uses 'physical_device' */
-		tmp = g_strdup_printf ("%s.physical_device", creator->category);
-		od = libhal_device_get_property_string (priv->hal_ctx, udi, tmp, NULL);
-		g_free (tmp);
+		usb_intf_udi = libhal_device_get_property_string (priv->hal_ctx, udi, "info.parent", NULL);
+		if (usb_intf_udi) {
+			od = libhal_device_get_property_string (priv->hal_ctx, usb_intf_udi, "info.parent", NULL);
+
+			/* Ensure the grandparent really is the "USB Device" */
+			if (od) {
+				tmp = hal_get_subsystem (priv->hal_ctx, od);
+				if (!tmp || strcmp (tmp, "usb_device")) {
+					libhal_free_string (od);
+					od = NULL;
+				}
+				if (tmp)
+					libhal_free_string (tmp);
+			}
+			libhal_free_string (usb_intf_udi);
+		}
 	}
+	libhal_free_string (subsys);
+	libhal_free_string (parent);
+
+	/* For non-USB devices, and as a fallback, just use the originating device
+	 * of the tty; though this might result in more than one modem being
+	 * detected by NM.
+	 */
+	if (!od)
+		od = hal_get_originating_device (priv->hal_ctx, udi, creator->category);
 
 	g_signal_emit (self, signals[UDI_ADDED], 0,
 	               udi,
@@ -693,6 +1310,13 @@ static void
 device_removed (LibHalContext *ctx, const char *udi)
 {
 	NMHalManager *self = NM_HAL_MANAGER (libhal_ctx_get_user_data (ctx));
+	DeferredModem *modem;
+
+	modem = deferred_modem_find (self, udi);
+	if (modem) {
+		deferred_modem_remove (self, modem);
+		deferred_modem_unref (modem);
+	}
 
 	g_signal_emit (self, signals[UDI_REMOVED], 0, udi);
 }
@@ -770,8 +1394,10 @@ killswitch_getpower_done (gpointer user_data)
 	if (priv->pending_polls > 0)
 		return;
 
-	if (priv->poll_rfkilled != priv->rfkilled) {
+	if (priv->poll_rfkilled != priv->rfkilled || !priv->killswitches_polled) {
 		priv->rfkilled = priv->poll_rfkilled;
+		nm_info ("Radio killswitches now %s radios",
+		         priv->rfkilled ? "block" : "allow");
 		g_signal_emit (self, signals[RFKILL_CHANGED], 0, priv->rfkilled);
 	}
 
@@ -781,6 +1407,8 @@ killswitch_getpower_done (gpointer user_data)
 	priv->killswitch_poll_id = g_timeout_add_seconds (RFKILL_POLL_FREQUENCY,
 	                                          poll_killswitches,
 	                                          self);
+
+	priv->killswitches_polled = TRUE;
 }
 
 static void 
@@ -1136,6 +1764,12 @@ dispose (GObject *object)
 	g_slist_foreach (priv->device_creators, destroy_creator, NULL);
 	g_slist_free (priv->device_creators);
 
+	g_slist_foreach (priv->deferred_modems, (GFunc) deferred_modem_unref, NULL);
+	g_slist_free (priv->deferred_modems);
+
+	if (priv->deferred_modem_id)
+		g_source_remove (priv->deferred_modem_id);
+
 	hal_deinit (self);
 
 	G_OBJECT_CLASS (nm_hal_manager_parent_class)->dispose (object);	
