@@ -48,8 +48,8 @@
 #include "nm-ifcfg-connection.h"
 #include "nm-inotify-helper.h"
 #include "shvar.h"
-
-#define IFCFG_DIR SYSCONFDIR"/sysconfig/network-scripts/"
+#include "writer.h"
+#include "utils.h"
 
 static void system_config_interface_init (NMSystemConfigInterface *system_config_interface_class);
 
@@ -198,43 +198,6 @@ read_one_connection (SCPluginIfcfg *plugin, const char *filename)
 	return connection;
 }
 
-static gboolean
-check_suffix (const char *basename, const char *tag)
-{
-	int len, tag_len;
-
-	g_return_val_if_fail (basename != NULL, TRUE);
-	g_return_val_if_fail (tag != NULL, TRUE);
-
-	len = strlen (basename);
-	tag_len = strlen (tag);
-	if ((len > tag_len) && !strcasecmp (basename + len - tag_len, tag))
-		return TRUE;
-	return FALSE;
-}
-
-static gboolean
-should_ignore_file (const char *filename)
-{
-	char *basename;
-	gboolean ignore = TRUE;
-
-	g_return_val_if_fail (filename != NULL, TRUE);
-
-	basename = g_path_get_basename (filename);
-	g_return_val_if_fail (basename != NULL, TRUE);
-
-	if (   !strncmp (basename, IFCFG_TAG, strlen (IFCFG_TAG))
-		&& !check_suffix (basename, BAK_TAG)
-		&& !check_suffix (basename, TILDE_TAG)
-		&& !check_suffix (basename, ORIG_TAG)
-		&& !check_suffix (basename, REJ_TAG))
-		ignore = FALSE;
-
-	g_free (basename);
-	return ignore;
-}
-
 static void
 read_connections (SCPluginIfcfg *plugin)
 {
@@ -248,7 +211,7 @@ read_connections (SCPluginIfcfg *plugin)
 		while ((item = g_dir_read_name (dir))) {
 			char *full_path;
 
-			if (should_ignore_file (item))
+			if (utils_should_ignore_file (item, TRUE))
 				continue;
 
 			full_path = g_build_filename (IFCFG_DIR, item, NULL);
@@ -294,9 +257,9 @@ connection_changed_handler (SCPluginIfcfg *plugin,
 		/* errors reading connection; remove it */
 		if (!ignore_error) {
 			PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    error: %s",
-			             error->message ? error->message : "(unknown)");
+			             (error && error->message) ? error->message : "(unknown)");
 		}
-		g_error_free (error);
+		g_clear_error (&error);
 
 		PLUGIN_PRINT (IFCFG_PLUGIN_NAME, "removed %s.", path);
 		*do_remove = TRUE;
@@ -337,7 +300,11 @@ connection_changed_handler (SCPluginIfcfg *plugin,
 		/* Only update if different */
 		if (!nm_connection_compare (new_wrapped, old_wrapped, NM_SETTING_COMPARE_FLAG_EXACT)) {
 			settings = nm_connection_to_hash (new_wrapped);
-			nm_exported_connection_update (NM_EXPORTED_CONNECTION (connection), settings, NULL);
+			if (!nm_ifcfg_connection_update (connection, settings, &error)) {
+				PLUGIN_WARN (IFCFG_PLUGIN_NAME, "    error updating: %s",
+				             (error && error->message) ? error->message : "(unknown)");
+				g_clear_error (&error);
+			}
 			g_hash_table_destroy (settings);
 		}
 
@@ -382,7 +349,6 @@ handle_connection_remove_or_new (SCPluginIfcfg *plugin,
 		}
 	}
 }
-
 static void
 dir_changed (GFileMonitor *monitor,
 		   GFile *file,
@@ -392,16 +358,20 @@ dir_changed (GFileMonitor *monitor,
 {
 	SCPluginIfcfg *plugin = SC_PLUGIN_IFCFG (user_data);
 	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
-	char *name;
+	char *path, *name;
 	NMIfcfgConnection *connection;
 	gboolean do_remove = FALSE, do_new = FALSE;
 
-	name = g_file_get_path (file);
-	if (should_ignore_file (name)) {
-		g_free (name);
+	path = g_file_get_path (file);
+	if (utils_should_ignore_file (path, FALSE)) {
+		g_free (path);
 		return;
 	}
 
+	/* Given any ifcfg, keys, or routes file, get the ifcfg file path */
+	name = utils_get_ifcfg_path (path);
+	g_free (path);
+
 	connection = g_hash_table_lookup (priv->connections, name);
 	if (!connection) {
 		do_new = TRUE;
@@ -435,7 +405,7 @@ setup_ifcfg_monitoring (SCPluginIfcfg *plugin)
 
 	priv->connections = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_object_unref);
 
-	file = g_file_new_for_path (IFCFG_DIR);
+	file = g_file_new_for_path (IFCFG_DIR "/");
 	monitor = g_file_monitor_directory (file, G_FILE_MONITOR_NONE, NULL, NULL);
 	g_object_unref (file);
 
@@ -472,6 +442,14 @@ get_connections (NMSystemConfigInterface *config)
 	return list;
 }
 
+static gboolean
+add_connection (NMSystemConfigInterface *config,
+                NMConnection *connection,
+                GError **error)
+{
+	return writer_new_connection (connection, IFCFG_DIR, NULL, error);
+}
+
 #define SC_NETWORK_FILE SYSCONFDIR"/sysconfig/network"
 
 static char *
@@ -479,6 +457,7 @@ plugin_get_hostname (SCPluginIfcfg *plugin)
 {
 	shvarFile *network;
 	char *hostname;
+	gboolean ignore_localhost;
 
 	network = svNewFile (SC_NETWORK_FILE);
 	if (!network) {
@@ -487,16 +466,18 @@ plugin_get_hostname (SCPluginIfcfg *plugin)
 	}
 
 	hostname = svGetValue (network, "HOSTNAME", FALSE);
-	svCloseFile (network);
-
-	/* Ignore a hostname of 'localhost' or 'localhost.localdomain' to preserve
-	 * 'network' service behavior.
-	 */
-	if (hostname && (!strcmp (hostname, "localhost") || !strcmp (hostname, "localhost.localdomain"))) {
-		g_free (hostname);
-		hostname = NULL;
+	ignore_localhost = svTrueValue (network, "NM_IGNORE_HOSTNAME_LOCALHOST", FALSE);
+	if (ignore_localhost) {
+		/* Ignore a hostname of 'localhost' or 'localhost.localdomain' to preserve
+		 * 'network' service behavior.
+		 */
+		if (hostname && (!strcmp (hostname, "localhost") || !strcmp (hostname, "localhost.localdomain"))) {
+			g_free (hostname);
+			hostname = NULL;
+		}
 	}
 
+	svCloseFile (network);
 	return hostname;
 }
 
@@ -512,7 +493,7 @@ plugin_set_hostname (SCPluginIfcfg *plugin, const char *hostname)
 		return FALSE;
 	}
 
-	svSetValue (network, "HOSTNAME", hostname);
+	svSetValue (network, "HOSTNAME", hostname, FALSE);
 	svWriteFile (network, 0644);
 	svCloseFile (network);
 
@@ -630,7 +611,7 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_string (value, IFCFG_PLUGIN_INFO);
 		break;
 	case NM_SYSTEM_CONFIG_INTERFACE_PROP_CAPABILITIES:
-		g_value_set_uint (value, NM_SYSTEM_CONFIG_INTERFACE_CAP_MODIFY_HOSTNAME);
+		g_value_set_uint (value, NM_SYSTEM_CONFIG_INTERFACE_CAP_MODIFY_CONNECTIONS | NM_SYSTEM_CONFIG_INTERFACE_CAP_MODIFY_HOSTNAME);
 		break;
 	case NM_SYSTEM_CONFIG_INTERFACE_PROP_HOSTNAME:
 		g_value_set_string (value, priv->hostname);
@@ -694,6 +675,7 @@ system_config_interface_init (NMSystemConfigInterface *system_config_interface_c
 {
 	/* interface implementation */
 	system_config_interface_class->get_connections = get_connections;
+	system_config_interface_class->add_connection = add_connection;
 	system_config_interface_class->get_unmanaged_devices = get_unmanaged_devices;
 	system_config_interface_class->init = init;
 }
